==> naifjs-0.44.0.tgz: package/lib/engine.js <==
const { banner } = require('../bin/lib/info')
const { getSessionId, setSessionId } = require('./sessionid')
const { validateRequestData } = require('./messaging/validateRequestData')
const { stopTimeout } = require('./timeouts')
const { setExternalResponse } = require('./externalResponse')

const { startTimeout } = require('./timeouts')
const { validateTimeout } = require('./messaging/validateTimeout')
const { validateResponseData, addWaitFlag } = require('./messaging/validateResponseData')
const { getExternalResponse } = require('./externalResponse')
const { warning } = require('./devlog')

// dialog handlers loader
const { Dialogs, loadUnits, printStates } = require('./unitsLoader')

// sessions
const { 
  isSessionActive,
  addDataToSession,
  getSession, 
  setSession, 
  delSession, 
  getSessionVar, 
  setSessionVar, 
  delSessionVar, 
  delSessionVars,
  initSessions, 
  dumpSessions 
} = require('./sessions') 
// eslint-disable-line no-unused-vars


//const TAGREGEXP = /[.:/]/
const TAGREGEXP = '.'


/*
 * dialog status as pointer to actual handler function
 * @private
 */
let unit
let state

/**
 * local pointer to external function to end
 * @private
 */
let externalEnd

/**
 * get current dialog stateid 
 * @return {String}
 */
function getCurrentUnitState() {
  return `${unit}.${state}`
}  


/**
 * set user variable
 *
 * @param {String} name of variable
 * @param {Object} value 
 * @param {String} unit name
 * @param {String} sessionid 
 *
 * @returns {Object} value
 *
 */
function setvar(name, value, unitid=unit, sessionid=getSessionId()) {
  // DEBUG
  //console.log(sessionid, unitid, name, value)
  return setSessionVar(sessionid, unitid, name, value)
}


/**
 * get user session variable
 *
 * @param {String} name name of variable
 * @param {String} unit name of unit
 * @param {String} sessionid 
 * @returns {Object} 
 *
 */
function getvar(name, unitid=unit, sessionid=getSessionId()) {
  // DEBUG
  //console.log(sessionid, unitid, name)
  return getSessionVar(sessionid, unitid, name)
}


/**
 * del user session variable
 *
 * @param {String} name name of variable
 * @param {String} unit name of unit
 * @param {String} sessionid 
 *
 */
function delvar(name, unitid=unit, sessionid=getSessionId()) {
  return delSessionVar(sessionid, unitid, name)
}


/**
 * del all user session variables
 *
 * @param {String} unit name of unit
 * @param {String} sessionid 
 *
 */
function delvars(unitid=unit, sessionid=getSessionId()) {
  return delSessionVars(sessionid, unitid)
}


/**
 * End of a conversation
 * Call an external handler
 *
 */
function end(sessionid=getSessionId()) {
  
  // set session not active
  delSession(sessionid)

  externalEnd(sessionid) 
}


/**
 * setup engine
 *
 * @public
 * @param {String} unitsFolder
 * @param {String} response / replyFunction
 * @param {String} stateid / default start state
 * @param {Boolean} [true] verbosity 
 *
 */
function up(unitsFolder, responseCallback, endCallback, sessionsFilename, verbose=true) {

  if (verbose) 
    console.info(banner())

  // load units
  build(unitsFolder, verbose)

  // assign external callback to reply 
  setExternalResponse(responseCallback)

  // assign external callback to end 
  externalEnd = endCallback

  // Load sessions in memory
  // if a valid filename is not passed, 
  // a sessions.json is created in the application directory
  initSessions(sessionsFilename ? sessionsFilename : unitsFolder+'/sessions.json')
  
}  


/**
 * shutdown engine
 *
 * @public
 *
 */
function down() {

  // dump sessions to disk
  dumpSessions()

} 


/**
 * load units
 * get the dialogs lookup map
 *
 * @private
 * @param {String} unitsFolder
 * @param {Boolean} verbose
 *
 */
function build(unitsFolder, verbose) {

  loadUnits(unitsFolder, verbose)
  
  if (verbose)
    printStates()

}


/**
 * start state by stateid
 *
 * @param {string} stateid unit.state string stateid 
 * @param {string} sessionid 
 * @param {AnyObject} argument (any object). It will passed to initial output state (by stateid) 
 */ 
function start(sessionid, stateid, argument) { 

  setSessionId(sessionid)

  // load the sessionid session assigning the initial state stateid
  exec(stateid, sessionid, argument)

}


/**
 * schedule and run an OUTPUT state 
 *
 * @param {string} stateid 
 * @param {string} sessionid user chat id. optional
 * @param {AnyObject} argument (any object). It will passed to initial output state (by stateid) 
 *
 * action
 *   set next dialog handler
 *   pass start argument as output state function argument 
 *   call the dialog 
 *
 */ 
function exec(stateid, sessionid=getSessionId(), argument) {

  next(stateid, sessionid)

  //
  // Call OUTPUT state function handler
  // pass the user id and the arg to the called function 
  Dialogs[unit].unit[state](sessionid, argument)

}


/**
 * schedule next INPUT state
 *
 * @param {string} stateid 
 * @param {string} sessionid user chat id. optional
 *
 * action
 *   prepare next dialog handler
 *   verify if unit.state exist looking at dialogs map
 *   store the status pointer
 *
 */ 
function next(stateid, sessionid=getSessionId()) {
  
  [unit, state] = stateid.split(TAGREGEXP)
  
  //
  // run-time state stateid validation
  //
  if ( !unit || !state ) 
    throw new Error('Wrong format. Please specify both unit and state with format: UnitName.StateName')
  
  if ( ! unitExist(unit) ) {
    const errorMsg = `dialog '${unit}' not found.`
    throw new Error(errorMsg)
  }  

  if ( ! stateExist(unit, state) ) {
    const errorMsg = `state '${state}' of dialog '${unit}' not found.`
    throw new Error(errorMsg)
  }  

  // Update session
  setSession(sessionid, unit, state)
}


/**
 * verify if unit exist in dialogs map
 *
 * @return {Boolean}
 *
 */ 
function unitExist(unit) {
  return Dialogs[unit] ? true : false
}


/**
 * verify if state exist in DialogName, looking up dialogs map
 *
 * @return {Boolean}
 *  
 */
function stateExist(unit, state) {
  return Dialogs[unit].states.includes(state)
}


/**
 * request
 * the user message trigger the the engine  
 *
 * @param {String} sessionid
 * @param {RequestDataObject | String} data - request data payload object
 *
 */
function request( sessionid, data ) { 

  // data could be a RequestDataObject ora String
  const validatedRequest = validateRequestData(data)

  setSessionId(sessionid)

  // clear any pending timeout action
  stopTimeout(sessionid)

  // Retrieve session data
  const session = getSession(sessionid)
  unit =  session.unit
  state = session.state

  // add request to session data
  addDataToSession(sessionid, { request: validatedRequest } )

  // DEBUG
  //console.log('calling dialog:', unit, ' state: ', state)

  //
  // Call INPUT state function handler
  // first argument is a validated RequestDataObject
  // second argument is the session id
  // https://stackoverflow.com/questions/8206453/call-function-by-string-name-on-node-js 
  // 
  Dialogs[unit].unit[state](validatedRequest, sessionid)
}


/**
 * response
 *
 * send a message to user, through an external function
 * set timeout callback if arg is specificed
 *
 * @param {String|ResponseDataObject} data    response data  Mandatory param
 * @param {TimeoutObject} timeoutDescriptor                  optional param 
 * @param {String} sessionid                                 optional param
 *
 * Where: 
 * @param {Object} TimeoutObject
 * @param {function} TimeoutObject.callback   function pointer
 * @param {ParamsObject} TimeoutObject.params parameters object 
 * @param {number} TimeoutObject.milliseconds number of timeout milliseconds 
 *
 *
 * @see https://www.jstips.co/en/javascript/passing-arguments-to-callback-functions/
 *
 * @return {function} externalResponse callback
 *
 */
function response(data, timeoutDescriptor, sessionid=getSessionId()) {

  const validatedData = validateResponseData(data)

  // start a timeout if argument 'timeoutDescriptor' is specificed
  if (timeoutDescriptor) {
    const { callback, params, milliseconds } = validateTimeout(timeoutDescriptor)

    startTimeout(callback, milliseconds, params, sessionid)
  }  

  // add the stateid to the ResponseDataObect
  validatedData.stateid = getCurrentUnitState()

  // add response to session data
  addDataToSession(sessionid, { response: validatedData } )

  //debug
  //console.info(validatedData)

  // run the callback, after a data object validation
  return getExternalResponse()(sessionid, validatedData)
}


/**
 * say
 * nowait response
 */ 
function say(data, sessionid) {
  return response(data, null, sessionid) 
}  


/**
 * ask
 * response that wait feedback from the user
 */ 
function ask(data, timeoutDescriptor, sessionid) {

  if ( !timeoutDescriptor )
    warning(`prompt(${JSON.stringify(data)}) without 'timeoutDescriptor' argument!`)

  // add wait flag to the data object
  return response(addWaitFlag(data), timeoutDescriptor, sessionid)
}  


// https://stackoverflow.com/a/3923258/1786393
module.exports = { 

  // out-dialog API
  getCurrentUnitState,
  currentUnitState: getCurrentUnitState, // alias

  up,
  start,

  request,
  run: request, // alias

  down,

  // in-dialog API
  exec,
  goout: exec, // alias
  setOutputState: exec, // alias
  gotoOutputState: exec, // alias
  nextOutputState: exec, // alias

  next,
  goto: next, // alias
  goinp: next, // alias
  setInputState: next, // alias
  gotoInputState: next, // alias
  nextInputState: next, // alias
  
  end,
  exit: end, // alias
  quit: end, // alias

  getvar,
  setvar,
  delvar,
  delvars,
  
  response,

  say, 
  reply: say, // alias

  ask,
  prompt: ask, // alias
  
  isSessionActive,
  isDialogActive: isSessionActive // alias
  
}

==> naifjs-0.45.0.tgz: package/lib/engine.js <==
const { banner } = require('../bin/lib/info')
const { getSessionId, setSessionId } = require('./sessionid')
const { getCurrentUnit, getCurrentState, getCurrentUnitState, setCurrentUnitState } = require('./stateid')
const { validateRequestData } = require('./messaging/validateRequestData')
const { stopTimeout } = require('./timeouts')
const { setExternalResponse } = require('./externalResponse')

// dialog handlers loader
const { Dialogs, loadUnits, printStates } = require('./unitsLoader')

// sessions
const { 
  isSessionActive,
  addDataToSession,
  getSession, 
  setSession, 
  delSession, 
  getSessionVar, 
  setSessionVar, 
  delSessionVar, 
  delSessionVars,
  initSessions, 
  dumpSessions 
} = require('./sessions') 
// eslint-disable-line no-unused-vars


/**
 * local pointer to external function to end
 * @private
 */
let externalEnd


/**
 * set user variable
 *
 * @param {String} name of variable
 * @param {Object} value 
 * @param {String} unit name
 * @param {String} sessionid 
 *
 * @returns {Object} value
 *
 */
function setvar(name, value, unitid=getCurrentUnit(), sessionid=getSessionId()) {
  // DEBUG
  //console.log(sessionid, unitid, name, value)
  return setSessionVar(sessionid, unitid, name, value)
}


/**
 * get user session variable
 *
 * @param {String} name name of variable
 * @param {String} unit name of unit
 * @param {String} sessionid 
 * @returns {Object} 
 *
 */
function getvar( name, unitid=getCurrentUnit(), sessionid=getSessionId() ) {
  // DEBUG
  //console.log(sessionid, unitid, name)
  return getSessionVar(sessionid, unitid, name)
}


/**
 * del user session variable
 *
 * @param {String} name name of variable
 * @param {String} unit name of unit
 * @param {String} sessionid 
 *
 */
function delvar( name, unitid=getCurrentUnit(), sessionid=getSessionId() ) {
  return delSessionVar(sessionid, unitid, name)
}


/**
 * del all user session variables
 *
 * @param {String} unit name of unit
 * @param {String} sessionid 
 *
 */
function delvars( unitid=getCurrentUnit(), sessionid=getSessionId() ) {
  return delSessionVars(sessionid, unitid)
}


/**
 * End of a conversation
 * Call an external handler
 *
 */
function end( sessionid=getSessionId() ) {
  
  // set session not active
  delSession(sessionid)

  externalEnd(sessionid) 
}


/**
 * up 
 * run the engine, loading session in memory 
 * 
 * @public
 *
 * @param {String} unitsFolder
 * @param {String} responseCallback
 * @param {String} sessionsFilename
 * @param {Boolean} [true] verbosity 
 *
 */
function up(unitsFolder, responseCallback, endCallback, sessionsFilename, verbose=true) {

  if (verbose) 
    console.info(banner())

  // load units
  build(unitsFolder, verbose)

  // assign external callback to reply 
  setExternalResponse(responseCallback)

  // assign external callback to end 
  externalEnd = endCallback

  // Load sessions in memory
  // if a valid filename is not passed, 
  // a sessions.json is created in the application directory
  initSessions(sessionsFilename ? sessionsFilename : unitsFolder+'/sessions.json')
  
}  


/**
 * shutdown the engine
 *
 * @public
 *
 */
function down() {

  // dump sessions to disk
  dumpSessions()

} 


/**
 * load units
 * get the dialogs lookup map
 *
 * @private
 * @param {String} unitsFolder
 * @param {Boolean} verbose
 *
 */
function build(unitsFolder, verbose) {

  loadUnits(unitsFolder, verbose)
  
  if (verbose)
    printStates()

}


/**
 * start a ne dialog session
 *
 * @public
 * @param {string} stateid unit.state string
 * @param {string} sessionid 
 * @param {AnyObject} argument (any object). It will passed to initial output state (by stateid) 
 */ 
function start(sessionid, stateid, argument) { 

  setSessionId(sessionid)

  // load the sessionid session assigning the initial stateid
  exec(stateid, sessionid, argument)

}


/**
 * schedule and run an OUTPUT NODE 
 *
 * @public
 * @param {string} stateid 
 * @param {string} [sessionid]
 * @param {AnyObject} argument (any object). It will passed to initial output state (by stateid) 
 *
 * action
 *   set next dialog handler
 *   pass start argument as output state function argument 
 *   call the dialog 
 *
 */ 
function exec(stateid, sessionid=getSessionId(), argument) {

  next(stateid, sessionid)

  //
  // Call OUTPUT state function handler
  // pass the user id and the arg to the called function 
  Dialogs[getCurrentUnit()].unit[getCurrentState()](sessionid, argument)

}


/**
 * schedule next INPUT NODE
 *
 * @public
 * @param {string} stateid 
 * @param {string} [sessionid]
 *
 * action
 *   prepare next dialog handler
 *   verify if unit.state exist looking at dialogs map
 *   store the status pointer
 *
 */ 
function next(stateid, sessionid=getSessionId()) {
  
  const [unit, state] = setCurrentUnitState(stateid)
  
  //
  // run-time state stateid validation
  //
  if ( !unit || !state ) 
    throw new Error('Wrong format. Please specify both unit and state with format: UnitName.StateName')
  
  if ( ! unitExist(unit) ) {
    const errorMsg = `dialog '${unit}' not found.`
    throw new Error(errorMsg)
  }  

  if ( ! stateExist(unit, state) ) {
    const errorMsg = `state '${state}' of dialog '${unit}' not found.`
    throw new Error(errorMsg)
  }  

  // Update session
  setSession(sessionid, unit, state)
}


/**
 * verify if unit exist in dialogs map
 *
 * @return {Boolean}
 *
 */ 
function unitExist(unit) {
  return Dialogs[unit] ? true : false
}


/**
 * verify if state exist in DialogName, looking up dialogs map
 *
 * @return {Boolean}
 *  
 */
function stateExist(unit, state) {
  return Dialogs[unit].states.includes(state)
}


/**
 * request
 * the user message trigger the the engine  
 *
 * @param {String} sessionid
 * @param {RequestDataObject | String} data - request data payload object
 *
 */
function request( sessionid, data ) { 

  // data could be a RequestDataObject ora String
  const validatedRequest = validateRequestData(data)

  setSessionId(sessionid)

  // clear any pending timeout action
  stopTimeout(sessionid)

  // Retrieve unit and state from session data
  const session = getSession(sessionid)

  const {unit, state} = session

  // add request to session data
  addDataToSession(sessionid, { request: validatedRequest } )

  // DEBUG
  //console.log('calling dialog:', unit, ' state: ', state)

  //
  // Call INPUT NODE function handler
  // first argument is a validated RequestDataObject
  // second argument is the session id
  // https://stackoverflow.com/questions/8206453/call-function-by-string-name-on-node-js 
  // 
  Dialogs[unit].unit[state](validatedRequest, sessionid)
}


// https://stackoverflow.com/a/3923258/1786393
module.exports = { 

  // out-dialog API
  getCurrentUnitState,
  currentUnitState: getCurrentUnitState, // alias

  up,
  start,

  request,
  run: request, // alias

  down,

  // in-dialog API
  exec,
  goout: exec, // alias
  setOutputState: exec, // alias
  gotoOutputState: exec, // alias
  nextOutputState: exec, // alias

  next,
  goto: next, // alias
  goinp: next, // alias
  setInputState: next, // alias
  gotoInputState: next, // alias
  nextInputState: next, // alias
  
  end,
  exit: end, // alias
  quit: end, // alias

  getvar,
  setvar,
  delvar,
  delvars,
  
  isSessionActive,
  isDialogActive: isSessionActive // alias
  
}
